# Copyright 2024 The jax_triton Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Edited in EasyDeL For caching kernels which doesn't work in jax-triton version
"""Module for calling Triton kernels from JAX."""

from __future__ import annotations

import copy
import dataclasses
import functools
import hashlib
import inspect
import os
import pickle
import pprint
import tempfile
import types
import zlib
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any, List, Protocol, Union

import jax
import jax.dlpack
import jax.numpy as jnp
import jaxlib
import numpy as np
from jax import tree_util
from jax._src import core, state, util
from jax._src.lib import gpu_triton as triton_kernel_call_lib
from jax._src.lib.mlir import ir
from jax.interpreters import mlir, xla

CAN_USE_TRITON = False
try:
	import triton
	import triton._C.libtriton as _triton
	import triton.backends.nvidia.compiler as cb
	import triton.language as tl
	from triton.compiler import code_generator as code_gen
	from triton.compiler import compiler as tc
	from triton.runtime import autotuner

	CAN_USE_TRITON = True
except ModuleNotFoundError:
	pass

try:
	import triton.backends.amd.compiler as hb
except ImportError:
	hb = None
	pass


def get_cache_dir() -> Path:
	home_dir = Path.home()
	app_name = "triton-compiled-kernels"
	if os.name == "nt":  # Windows
		cache_dir = (
			Path(os.getenv("LOCALAPPDATA", home_dir / "AppData" / "Local")) / app_name
		)
	elif os.name == "posix":  # Linux and macOS
		if "darwin" in os.sys.platform:  # macOS
			cache_dir = home_dir / "Library" / "Caches" / app_name
		else:  # Linux
			cache_dir = home_dir / ".cache" / app_name
	else:
		cache_dir = home_dir / ".cache" / app_name
	cache_dir.mkdir(parents=True, exist_ok=True)
	return cache_dir


_JAX_TRITON_DUMP_DIR = Path(os.environ.get("JAX_TRITON_DUMP_DIR", get_cache_dir()))
_CACHE_TRITON_KERNELS = os.environ.get("CACHE_TRITON_KERNELS", "true") in [
	"1",
	"on",
	"true",
	"yes",
]
map, unsafe_map = util.safe_map, map
zip, unsafe_zip = util.safe_zip, zip


_JAX_TO_TRITON_TYPE_MAP = {
	jnp.dtype("bfloat16"): "bf16",
	jnp.dtype("float64"): "fp64",
	jnp.dtype("float32"): "fp32",
	jnp.dtype("float16"): "fp16",
	jnp.dtype("float8_e4m3fn"): "fp8e4nv",
	jnp.dtype("float8_e5m2"): "fp8e5",
	jnp.dtype("float8_e4m3fnuz"): "fp8e4b8",
	jnp.dtype("float8_e5m2fnuz"): "fp8e5b16",
	jnp.dtype("int64"): "i64",
	jnp.dtype("int32"): "i32",
	jnp.dtype("int16"): "i16",
	jnp.dtype("int8"): "i8",
	jnp.dtype("uint64"): "u64",
	jnp.dtype("uint32"): "u32",
	jnp.dtype("uint16"): "u16",
	jnp.dtype("uint8"): "u8",
	# Triton defines a 'B' type, which is an alias for both i1 and bool.
	jnp.dtype("bool"): "B",
}

Grid = Union[int, tuple[int], tuple[int, int], tuple[int, int, int]]
GridOrLambda = Union[Grid, Callable[[dict[str, Any]], Grid]]


def normalize_grid(grid: GridOrLambda, metaparams) -> tuple[int, int, int]:
	if callable(grid):
		grid = grid(metaparams)
	if isinstance(grid, int):
		grid = (grid,)
	elif len(grid) > 3:
		raise ValueError("`grid` should have three or fewer dimensions.")
	return tuple(grid) + (1,) * (3 - len(grid))


def avals_to_layouts(avals):
	return [list(reversed(range(aval.ndim))) for aval in avals]


def get_triton_type(obj: Any) -> str:
	if isinstance(obj, (jax.core.ShapedArray, state.AbstractRef)):
		return f"*{_JAX_TO_TRITON_TYPE_MAP[obj.dtype]}"
	if isinstance(obj, tl.constexpr):
		obj = obj.value
	if isinstance(obj, int):
		if -(2**31) <= obj < 2**31:
			return "i32"
		elif 2**31 <= obj < 2**32:
			return "u32"
		elif -(2**63) <= obj < 2**63:
			return "i64"
		elif 2**63 <= obj < 2**64:
			return "u64"
		else:
			raise ValueError(f"integer overflow representing {obj}")
	if isinstance(obj, float):
		return "fp64"
	if isinstance(obj, np.float32):
		return "fp32"
	if isinstance(obj, bool):
		return "B"
	if isinstance(obj, str):
		return "str"
	raise NotImplementedError(f"could not compute type name for {obj}: {type(obj)}")


triton_kernel_call_p = jax.core.Primitive("triton_kernel_call")
triton_kernel_call_p.multiple_results = True
triton_kernel_call_p.def_impl(
	functools.partial(xla.apply_primitive, triton_kernel_call_p)
)


@triton_kernel_call_p.def_abstract_eval
def triton_kernel_call_abstract_eval(*_, out_shapes, **__):
	return [
		core.ShapedArray(out_shape.shape, out_shape.dtype) for out_shape in out_shapes
	]


def aval_size_bytes(aval):
	return np.dtype(aval.dtype).itemsize * aval.size


def get_cuda_backend(device, compute_capability):
	target = cb.GPUTarget("cuda", compute_capability, 32)
	backend = cb.CUDABackend(target)
	return backend


def get_hip_backend(device, compute_capability):
	arch = triton_kernel_call_lib.get_arch_details(device)
	arch = arch.split(":")[0]
	target = hb.GPUTarget("hip", arch, 64)
	backend = hb.HIPBackend(target)
	return backend


@dataclasses.dataclass
class CompilationResult:
	binary: str
	name: str
	shared_mem_bytes: int
	cluster_dims: tuple
	ttgir: str | None
	llir: str | None


def compile_ttir_inplace(
	ttir,
	backend: List[Union[cb.CUDABackend, hb.HIPBackend]],  # type:ignore
	options: List[Union[cb.CUDAOptions, hb.HIPOptions]],  # type:ignore
	compute_capability,
	platform,
):
	if platform == "cuda":
		return compile_ttir_to_ptx_inplace(
			ttir,
			backend,
			options,
			compute_capability,
		)

	elif platform == "rocm":
		return compile_ttir_to_hsaco_inplace(
			ttir,
			backend,
			options,
			compute_capability,
		)
	else:
		raise ValueError("Unsupported device.")


def compile_ttir_to_ptx_inplace(
	ttir,
	cuda_backend: cb.CUDABackend,
	cuda_options: cb.CUDAOptions,
	compute_capability,
) -> CompilationResult:
	if cuda_options.debug:
		print(ttir)
	try:
		metadata = {}
		opt_ttir = cuda_backend.make_ttir(ttir, metadata, cuda_options)
		ttgir = cuda_backend.make_ttgir(
			opt_ttir,
			metadata,
			cuda_options,
			compute_capability,
		)
	except RuntimeError as e:
		ttir.dump()
		raise ValueError("TTIR->TTGIR pass failed!") from e
	if cuda_options.debug:
		print(ttgir)
	try:
		llir = cuda_backend.make_llir(
			ttgir,
			metadata,
			cuda_options,
			compute_capability,
		)
	except RuntimeError as e:
		ttgir.dump()
		raise ValueError("TTGIR->LLIR pass failed!") from e
	shared_mem_bytes = metadata["shared"]
	if cuda_options.debug:
		print(llir)
	ptx = cuda_backend.make_ptx(
		llir,
		metadata,
		cuda_options,
		compute_capability,
	)
	if cuda_options.debug:
		print(ptx)
	name = metadata["name"]
	cluster_dims = metadata["cluster_dims"]
	ttgir = str(ttgir)
	llir = str(llir)
	return CompilationResult(
		binary=ptx,
		name=name,
		shared_mem_bytes=shared_mem_bytes,
		cluster_dims=cluster_dims,
		ttgir=ttgir,
		llir=llir,
	)


def compile_ttir_to_hsaco_inplace(
	ttir,
	hip_backend: hb.HIPBackend,  # type:ignore
	hip_options: hb.HIPOptions,  # type:ignore
	compute_capability,
) -> CompilationResult:
	if hip_options.debug:
		print(ttir)
	try:
		metadata = {}
		opt_ttir = hip_backend.make_ttir(ttir, metadata, hip_options)
		ttgir = hip_backend.make_ttgir(opt_ttir, metadata, hip_options)
	except RuntimeError as e:
		ttir.dump()
		raise ValueError("TTIR->TTGIR pass failed!") from e
	if hip_options.debug:
		print(ttgir)
	try:
		llir = hip_backend.make_llir(ttgir, metadata, hip_options)
	except RuntimeError as e:
		ttgir.dump()
		raise ValueError("TTGIR->LLIR pass failed!") from e
	shared_mem_bytes = metadata["shared"]
	if hip_options.debug:
		print(llir)

	amdgcn = hip_backend.make_amdgcn(llir, metadata, hip_options)
	hsaco = hip_backend.make_hsaco(amdgcn, metadata, hip_options)

	name = metadata["name"]
	ttgir = str(ttgir)
	llir = str(llir)
	cluster_dims = (0, 0, 0)
	fd, hsaco_path = tempfile.mkstemp()
	with os.fdopen(fd, "wb") as f:
		f.write(hsaco)
	return CompilationResult(
		binary=hsaco_path,
		name=name,
		shared_mem_bytes=shared_mem_bytes,
		cluster_dims=cluster_dims,
		ttgir=ttgir,
		llir=llir,
	)


_COMPILED_KERNEL_CACHE = {}


def get_or_create_triton_kernel(
	backend_init_func,
	platform,
	fn,
	arg_dtypes,
	scalar_args,
	*,
	num_warps,
	num_stages,
	num_ctas,
	compute_capability,
	enable_fp_fusion,
	metaparams,
	dump: bool,
) -> tuple[triton_kernel_call_lib.TritonKernel, Any]:  # type:ignore
	if num_warps is None:
		num_warps = 4
	if num_stages is None:
		num_stages = 3
	device = 0
	if compute_capability is None:
		compute_capability = triton_kernel_call_lib.get_compute_capability(device)
	if num_ctas > 1 and compute_capability < 90:
		raise ValueError("num_ctas > 1 unsupported before Hopper.")

	signature = {fn.arg_names[i]: v for i, v in enumerate(arg_dtypes)}

	mock_torch_tensor = types.SimpleNamespace(data_ptr=lambda: 16)
	args_for_specialization_attr = [mock_torch_tensor] * len(arg_dtypes)
	for i, _, v in scalar_args:
		args_for_specialization_attr[i] = v
	specialization_attr = fn._get_config(*args_for_specialization_attr)  # pylint: disable=protected-access

	constants = {k: v for k, v in metaparams.items()}
	constants.update({k: None for _, k, v in scalar_args if v is None})
	constants.update({fn.arg_names[i]: 1 for i in specialization_attr.equal_to_1})

	cache_key = (
		fn,
		tuple(signature.items()),
		tuple(vars(specialization_attr).values()),
		tuple(constants.items()),
		num_warps,
		num_stages,
		num_ctas,
		compute_capability,
		enable_fp_fusion,
	)
	kernel = _COMPILED_KERNEL_CACHE.get(cache_key)

	kernel_hash = hashlib.sha256(str(cache_key).encode("utf-8")).hexdigest()

	opts = {
		"num_warps": num_warps,
		"num_stages": num_stages,
		"num_ctas": num_ctas,
		"optimize_epilogue": False,
		"debug": dump,
		"enable_fp_fusion": enable_fp_fusion,
	}
	if kernel is None and not Path(_JAX_TRITON_DUMP_DIR / kernel_hash).exists():
		backend = backend_init_func(device, compute_capability)
		options = backend.parse_options(opts)

		if _CACHE_TRITON_KERNELS:
			os.makedirs(f"{_JAX_TRITON_DUMP_DIR}/{kernel_hash}")
			with open(f"{_JAX_TRITON_DUMP_DIR}/{kernel_hash}/config", "w") as f:
				pprint.pprint(cache_key, stream=f)
				pprint.pprint(options, stream=f)

		context = _triton.ir.context()
		_triton.ir.load_dialects(context)
		backend.load_dialects(context)
		codegen_fns = backend.get_codegen_implementation()

		module = (
			code_gen.ast_to_ttir(
				fn,
				specialization=tc.ASTSource(
					fn,
					constants=constants,
					signature=signature,
					attrs=specialization_attr,
				),
				options=options,
				codegen_fns=codegen_fns,
				context=context,
				module_map=backend.get_module_map(),
			)
			if "module_map" in inspect.getfullargspec(code_gen.ast_to_ttir).args
			# Triton changes ASTSource.ast_to_ttir to include module_map. Handle
			# backward compatibility here.
			else code_gen.ast_to_ttir(
				fn,
				specialization=tc.ASTSource(
					fn,
					constants=constants,
					signature=signature,
					attrs=specialization_attr,
				),
				options=options,
				codegen_fns=codegen_fns,
				context=context,
			)
		)
		ttir = str(module)

		compilation_result = compile_ttir_inplace(
			module,
			backend,
			options,
			compute_capability,
			platform,
		)
		if _CACHE_TRITON_KERNELS:
			(_JAX_TRITON_DUMP_DIR / kernel_hash).mkdir(parents=True, exist_ok=True)
			pickle.dump(
				(compilation_result, ttir),
				open(_JAX_TRITON_DUMP_DIR / kernel_hash / "compilation_result", "wb"),
			)
		kernel_name = compilation_result.name
		kernel = triton_kernel_call_lib.TritonKernel(
			kernel_name,
			num_warps,
			compilation_result.shared_mem_bytes,
			compilation_result.binary,
			ttir,
			compute_capability,
			*compilation_result.cluster_dims,
		)

		_COMPILED_KERNEL_CACHE[cache_key] = kernel
	elif Path(_JAX_TRITON_DUMP_DIR / kernel_hash).exists():
		compilation_result, ttir = pickle.load(
			open(
				_JAX_TRITON_DUMP_DIR / kernel_hash / "compilation_result",
				"rb",
			)
		)
		kernel = triton_kernel_call_lib.TritonKernel(
			compilation_result.name,
			num_warps,
			compilation_result.shared_mem_bytes,
			compilation_result.binary,
			ttir,
			compute_capability,
			*compilation_result.cluster_dims,
		)

	return kernel, specialization_attr


def triton_kernel_call_lowering(
	backend_init_func,
	ctx,
	*array_args,
	fn,
	scalar_args,
	name,
	custom_call_target_name,
	out_shapes,
	grid,
	num_warps,
	num_stages,
	num_ctas,
	compute_capability,
	enable_fp_fusion,
	input_output_aliases,
	zeroed_outputs,
	debug,
	serialized_metadata,
	**metaparams,
):
	if jaxlib.version.__version_info__ < (0, 3, 22) and input_output_aliases:
		raise NotImplementedError(
			"`input_output_aliases` only supported on `jaxlib>=0.3.22"
		)

	kernel_call_name = name
	args = list(ctx.avals_in)
	arg_dtypes = list(map(get_triton_type, ctx.avals_in))
	for idx, dtype, v in scalar_args:
		args.insert(idx, v)
		arg_dtypes.insert(idx, dtype)
	args.extend(ctx.avals_out)
	arg_dtypes.extend(map(get_triton_type, ctx.avals_out))
	named_args = dict(unsafe_zip(fn.arg_names, args))

	if isinstance(fn, autotuner.Autotuner):
		prev_early_config_prune_fn = fn.early_config_prune

		def prune_configs(configs, named_args, **kwargs):
			pruned_configs = []
			for config in configs:
				if config.pre_hook is not None:
					raise NotImplementedError("`pre_hook` is not supported")

				if all(config.kwargs.get(k, v) == v for k, v in metaparams.items()):
					pruned_configs.append(config)
			if prev_early_config_prune_fn is not None:
				pruned_configs = prev_early_config_prune_fn(pruned_configs, named_args)
			return pruned_configs

		fn.early_config_prune = prune_configs
		fn.nargs = named_args
		configs = fn.prune_configs(metaparams)
		fn = fn.fn
	else:
		config = triton.Config(
			{},
			num_warps=num_warps,
			num_stages=num_stages,
			num_ctas=num_ctas,
		)
		configs = [config]

	if isinstance(fn, autotuner.Heuristics):
		updated_configs = []
		for config in configs:
			kwargs = config.kwargs.copy()
			for name, heuristic in fn.values.items():
				kwargs[name] = heuristic({**named_args, **metaparams, **kwargs})
			updated_config = copy.copy(config)
			updated_config.kwargs = kwargs
			updated_configs.append(updated_config)
		configs = updated_configs
		fn = fn.fn

	if not isinstance(fn, triton.JITFunction):
		raise ValueError(
			"`kernel` must be a Triton `JITFunction`, `Heuristics` or `Autotuner`."
		)

	outputs_offset = len(ctx.avals_in) + len(scalar_args)
	config_params = []
	for config in configs:
		config_metaparams = {**metaparams, **config.kwargs}
		config_grid = normalize_grid(grid, config_metaparams)

		config_zeroed_outputs = zeroed_outputs
		if callable(zeroed_outputs):
			config_zeroed_outputs = config_zeroed_outputs(config_metaparams)

		zeroed_params_with_sizes = {
			i + outputs_offset: aval_size_bytes(ctx.avals_out[i])
			for i in sorted(config_zeroed_outputs)
		}

		config_params.append(
			dict(
				metaparams=tuple(sorted(config_metaparams.items())),
				num_warps=config.num_warps,
				num_stages=config.num_stages,
				num_ctas=config.num_ctas,
				grid=config_grid,
				zeroed_params_with_sizes=tuple(zeroed_params_with_sizes.items()),
			)
		)

	kernel_calls = []
	for params in config_params:
		kernel, specialization_attr = get_or_create_triton_kernel(
			backend_init_func,
			ctx.module_context.platforms[0],
			fn,
			arg_dtypes,
			scalar_args,
			num_warps=params["num_warps"],
			num_stages=params["num_stages"],
			num_ctas=params["num_ctas"],
			compute_capability=compute_capability,
			enable_fp_fusion=enable_fp_fusion,
			metaparams=dict(params["metaparams"]),
			dump=debug,
		)

		kernel_params = []
		zeroed_params_with_sizes = dict(params["zeroed_params_with_sizes"])
		for i, (arg, dtype) in enumerate(zip(args, arg_dtypes)):
			if isinstance(arg, core.ShapedArray):
				kernel_params.append(
					triton_kernel_call_lib.create_array_parameter(
						zeroed_params_with_sizes.get(i, 0),
						16 if (i in specialization_attr.divisible_by_16) else 0,
					)
				)
			elif i not in specialization_attr.equal_to_1:
				kernel_params.append(triton_kernel_call_lib.create_scalar_parameter(arg, dtype))

		kernel_calls.append(
			triton_kernel_call_lib.TritonKernelCall(
				kernel,
				params["grid"][0],
				params["grid"][1],
				params["grid"][2],
				kernel_params,
			)
		)

	if len(kernel_calls) > 1:
		named_scalar_args = {fn.arg_names[i]: v for i, _, v in scalar_args}
		input_output_aliases_with_sizes = tuple(
			(input_idx, output_idx, aval_size_bytes(ctx.avals_in[input_idx]))
			for input_idx, output_idx in input_output_aliases
		)
		kernel_call = triton_kernel_call_lib.TritonAutotunedKernelCall(
			f"{kernel_call_name} ({fn.fn.__name__}) {named_scalar_args}",
			[(call, str(config)) for call, config in zip(kernel_calls, configs)],
			input_output_aliases_with_sizes,
		)
	else:
		kernel_call = kernel_calls[0]

	out_types = [
		ir.RankedTensorType.get(shape.shape, mlir.dtype_to_ir_type(shape.dtype))
		for shape in out_shapes
	]
	if jaxlib.version.__version_info__ >= (0, 4, 15):
		call_proto = kernel_call.to_proto(kernel_call_name, serialized_metadata)
	else:
		call_proto = kernel_call.to_proto(serialized_metadata)
	return jaxlib.hlo_helpers.custom_call(
		call_target_name=custom_call_target_name,
		result_types=out_types,
		operands=array_args,
		backend_config=zlib.compress(call_proto),
		operand_layouts=avals_to_layouts(ctx.avals_in),
		result_layouts=avals_to_layouts(ctx.avals_out),
		operand_output_aliases=dict(input_output_aliases),
	).results


mlir.register_lowering(
	triton_kernel_call_p,
	functools.partial(triton_kernel_call_lowering, get_cuda_backend),
	platform="cuda",
)

mlir.register_lowering(
	triton_kernel_call_p,
	functools.partial(triton_kernel_call_lowering, get_hip_backend),
	platform="rocm",
)


class ShapeDtype(Protocol):
	@property
	def shape(self) -> tuple[int, ...]: ...

	@property
	def dtype(self) -> np.dtype: ...


def triton_call(
	*args: jax.Array | bool | int | float | np.float32,
	kernel: triton.JITFunction,
	out_shape: ShapeDtype | Sequence[ShapeDtype],
	grid: GridOrLambda,
	name: str = "",
	custom_call_target_name: str = "triton_kernel_call",
	num_warps: int | None = None,
	num_stages: int | None = None,
	num_ctas: int = 1,
	compute_capability: int | None = None,
	enable_fp_fusion: bool = True,
	input_output_aliases: dict[int, int] | None = None,
	zeroed_outputs: (Sequence[int] | Callable[[dict[str, Any]], Sequence[int]]) = (),
	debug: bool = False,
	serialized_metadata: bytes = b"",
	**metaparams: Any,
) -> Any:
	"""Calls a Triton kernel with `jax.Array` arguments.
	Args:
	  *args: Inputs for the Triton kernel.
	  kernel: A Triton kernel (e.g. a function decorated with `triton.jit`). All
	    static values should be annotated with `triton.language.constexpr`.
	  out_shape: A `jax.ShapeDtypeStruct` (or something that has `.shape` and
	    `.dtype` attributes) or a sequence thereof that specify the output(s) of
	    the kernel. Pointers for each of the `jax.ShapeDtypeStruct`s in
	    `out_shape` will be passed into `kernel` following the input parameters.
	  grid: An integer, tuple of up to 3 integers, or a function that returns a
	    tuple of up to 3 integers. When `grid` is an integer, `kernel` is
	    invocated in `grid`-many parallel executions. When `grid` is a sequence of
	    integers, `kernel` is launched in a `prod(grid)`-many parallel execution.
	    When `grid` is a function, it is passed `**metaparams` and should return a
	    tuple of up to 3 integers.
	  input_output_aliases: A dictionary mapping input argument indices to output
	    indices. Providing a mapping will alias the corresponding buffers.
	  zeroed_outputs: A sequence of indices, or a function returning a sequence of
	    indices, for outputs that should be zeroed before the kernel is launched.
	  num_warps: The number of warps used to execute the Triton kernel.
	  num_stages: The number of stages emitted by the Triton compiler.
	  num_ctas: The size of thread blocks per cluster to be used on GPUs with
	    compute capabilities >= 9.0. It must be less or equal to 8.
	  debug: Prints out intermediate IRs if True for debugging purposes.
	  serialized_metadata: Arbitrary metadata that will be added into the
	    serialized kernel call.
	  **metaparams: Additional keyword arguments that will be provided to a `grid`
	    (if it is a function) and to the Triton kernel as `constexpr` arguments.

	Returns:
	  Outputs from the Triton kernel.
	"""
	if not CAN_USE_TRITON:
		raise ValueError("`triton_call` is only available when `triton` is installed.")
	out_shape = tree_util.tree_map(
		lambda a: jax.ShapeDtypeStruct(a.shape, a.dtype), out_shape
	)
	flat_args, _ = tree_util.tree_flatten(args)
	flat_out_shapes, out_tree = tree_util.tree_flatten(out_shape)

	array_args = []
	scalar_args = []
	for i, arg in enumerate(flat_args):
		if isinstance(arg, (bool, int, float)):
			scalar_args.append((i, get_triton_type(arg), arg))
		elif isinstance(arg, np.float32):
			scalar_args.append((i, get_triton_type(arg), float(arg)))
		else:
			array_args.append(arg)

	if input_output_aliases is None:
		input_output_aliases = {}

	out_flat = triton_kernel_call_p.bind(
		*array_args,
		fn=kernel,
		scalar_args=tuple(scalar_args),
		name=name,
		custom_call_target_name=custom_call_target_name,
		out_shapes=tuple(flat_out_shapes),
		grid=grid,
		num_warps=num_warps,
		num_stages=num_stages,
		num_ctas=num_ctas,
		compute_capability=compute_capability,
		enable_fp_fusion=enable_fp_fusion,
		input_output_aliases=tuple(input_output_aliases.items()),
		zeroed_outputs=zeroed_outputs,
		debug=debug,
		serialized_metadata=serialized_metadata,
		**metaparams,
	)
	return tree_util.tree_unflatten(out_tree, out_flat)



"""
__init__ module for custom DataTypes created by fjformer.core
"""

from fjformer.dtypes.arraynf4 import ArrayNF4 as ArrayNF4
from fjformer.dtypes.array8bit import Array8Bit as Array8Bit
from fjformer.dtypes.array8lt import Array8Lt as Array8Lt
from fjformer.dtypes.a8q import A8Q as A8Q
from fjformer.dtypes.a4q import A4Q as A4Q



import chex
import jax
from jax import numpy as jnp
from jax._src.flatten_util import ravel_pytree


def global_norm(tree: chex.ArrayTree) -> chex.Array:
    """
    Computes the global norm of a PyTree of arrays.

    The global norm is the square root of the sum of squares of all elements in the
    PyTree.

    Args:
        tree: The PyTree of arrays.

    Returns:
        The global norm as a scalar JAX array.
    """
    squared = jax.tree_util.tree_map(lambda x: jnp.sum(jnp.square(x)), tree)
    flattened, _ = ravel_pytree(squared)
    return jnp.sqrt(jnp.sum(flattened))


def average_metrics(metrics: list[chex.ArrayTree]) -> chex.ArrayTree:
    """
    Averages a list of metrics, each represented as a PyTree.

    This function is useful for averaging metrics across multiple devices or steps.

    Args:
        metrics: A list of PyTrees representing the metrics.

    Returns:
        A PyTree with the same structure as the input metrics, but with each leaf
        node replaced by the average of the corresponding leaf nodes across the
        input metrics.
    """
    return jax.tree_util.tree_map(lambda *args: jnp.mean(jnp.stack(args)), *metrics)


def transpose(array: chex.Array, dim0: int, dim1: int) -> chex.Array:
    """
    Transposes two dimensions of a JAX array.

    This function provides a convenient way to swap two dimensions of an array using
    negative indexing for dimensions.

    Args:
        array: The input array.
        dim0: The first dimension to transpose. Can be negative.
        dim1: The second dimension to transpose. Can be negative.

    Returns:
        The transposed array.
    """
    dim0 = dim0 if dim0 > 0 else array.ndim + dim0
    dim1 = dim1 if dim1 > 0 else array.ndim + dim1
    perm = list(range(array.ndim))
    perm[dim0], perm[dim1] = perm[dim1], perm[dim0]
    return jnp.transpose(array, perm)


def fused_softmax(x: chex.Array, axis: int = -1) -> chex.Array:
    """
    Computes a numerically stable softmax using a fused implementation.

    This function computes the softmax by first calculating the log-softmax and
    then exponentiating the result. This approach is more numerically stable than
    directly computing the softmax.

    Args:
        x: The input array.
        axis: The axis along which to compute the softmax.

    Returns:
        The softmax of the input array.
    """
    return jnp.exp(jax.nn.log_softmax(x, axis=axis))



# Copyright 2023 The EASYDEL Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation by @erfanzar,
# with a few bug fixes and adjustments.

from functools import partial
from typing import Optional

import jax
import jax.interpreters
import jax.interpreters.pxla
import jax.random
from fjformer import GenerateRNG
from jax import numpy as jnp
from jax.experimental import pallas as pl

PLATFORM = jax.extend.backend.get_backend().platform
INTERPRET = PLATFORM == "cpu"
rng = GenerateRNG()

# GPU KERNEL


def _gpu_matmul_kernel_fwd(
	a_ref,
	b_ref,
	o_ref,
	*,
	blocksize_m,
	blocksize_k,
	blocksize_n,
	prod_dtype,
	precision,
):
	row_id, col_id = pl.program_id(0), pl.program_id(1)

	m, n = a_ref.shape[0], b_ref.shape[1]
	col_slice = pl.dslice(col_id * blocksize_n, blocksize_n)
	row_slice = pl.dslice(row_id * blocksize_m, blocksize_m)
	a_mask_i = (row_id * blocksize_m + jnp.arange(blocksize_m) < a_ref.shape[0])[:, None]
	b_mask_j = (col_id * blocksize_n + jnp.arange(blocksize_n) < b_ref.shape[1])[None, :]
	ij = jnp.arange(blocksize_k)

	def body(start_i, carry_i):
		o_p = carry_i
		a_mask_j = (start_i * blocksize_k + ij < jnp.arange(blocksize_k))[None, :] < m
		b_mask_i = (start_i * blocksize_k + ij < jnp.arange(blocksize_k))[:, None] < n
		inner_slice = pl.dslice(start_i * blocksize_k, blocksize_k)

		a = pl.load(
			a_ref,
			(row_slice, inner_slice),
			mask=a_mask_i & a_mask_j,
			other=0.0,
		)

		b = pl.load(
			b_ref,
			(inner_slice, col_slice),
			mask=b_mask_i & b_mask_j,
			other=0.0,
		)
		return jnp.dot(a, b, precision=precision) + o_p

	o = jax.lax.fori_loop(
		0,
		pl.cdiv(b_ref.shape[0], blocksize_k),
		body,
		jnp.zeros((blocksize_m, blocksize_n), dtype=prod_dtype),
	)
	omi = (row_id * blocksize_m + jnp.arange(blocksize_m) < o_ref.shape[0])[:, None]
	omj = (col_id * blocksize_n + jnp.arange(blocksize_n) < o_ref.shape[1])[None, :]
	o_mask = omi & omj
	pl.store(
		o_ref,
		(
			pl.dslice(row_id * blocksize_m, blocksize_m),
			pl.dslice(col_id * blocksize_n, blocksize_n),
		),
		o.astype(o_ref.dtype),
		mask=o_mask,
	)


def _get_compiler_params(
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	dtype: jnp.dtype,
):
	params = None
	platform = jax.extend.backend.get_backend().platform
	dtype_size = jnp.dtype(dtype).itemsize
	if platform == "gpu":
		dtype_size = jnp.dtype(dtype).itemsize

		# Calculate shared memory usage
		shared_mem = 2 * blocksize_k * (blocksize_m + blocksize_n) * dtype_size

		# Improved num_warps calculation
		num_threads = blocksize_m * blocksize_n
		num_warps = min(max(num_threads // 32, 1), 8)

		# Adjust num_warps based on shared memory usage and block size
		if shared_mem > 32768:  # 32 KB shared memory limit for many GPUs
			num_warps = min(num_warps, 4)
		elif max(blocksize_m, blocksize_n) >= 128:
			num_warps = max(num_warps, 4)

		# num_stages calculation (as before)
		l2_cache_size = 1024 * 1024
		working_set_size = shared_mem * 2
		num_stages = min(max(l2_cache_size // working_set_size, 2), 5)

		if blocksize_k <= 32:
			num_stages = min(num_stages, 2)
		elif blocksize_k <= 64:
			num_stages = min(num_stages, 3)

		params = dict(triton=dict(num_stages=num_stages, num_warps=num_warps))

	return params


def get_best_block_size(A, B):
	# A is assumed to be of shape (m, k) and B is of shape (k, n)
	m, k = A.shape[0], A.shape[1]
	n = B.shape[1]

	# Initialize block sizes
	bm, bk, bn = 16, 16, 16

	# Adjust block size for m
	if m > 1000:
		bm = 32
	if m > 2000:
		bm = 64
	if m > 8000:
		bm = 128

	# Adjust block size for k
	if k > 1000:
		bk = 32
	if k > 2000:
		bk = 64
	if k > 8000:
		bk = 128

	# Adjust block size for n
	if n > 1000:
		bn = 32  # bn maxes out at 32

	return bm, bk, bn


@partial(
	jax.jit,
	static_argnames=[
		"blocksize_m",
		"blocksize_k",
		"blocksize_n",
		"prod_dtype",
		"precision",
	],
)
def _call_gpu_matmul_kernel_fwd(
	A: jax.Array,
	B: jax.Array,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	prod_dtype: jnp.dtype,
	precision: jax.lax.PrecisionLike = None,
):
	# A(mk)@B(kn)=C(mn)
	assert A.ndim == 2 and B.ndim == 2, f"got {A.shape=} and {B.shape=}"
	assert (
		A.shape[1] == B.shape[0]
	), f"matmul can't be operated with these shapes {A.shape=} {B.shape=} "
	m, n = A.shape[0], B.shape[1]
	pbm, pbk, pbn = get_best_block_size(A, B)

	blocksize_m = blocksize_m or pbm
	blocksize_n = blocksize_n or pbn
	blocksize_k = blocksize_k or pbk

	grid = (pl.cdiv(m, blocksize_m), pl.cdiv(n, blocksize_n))
	in_specs = [
		pl.BlockSpec(A.shape, lambda *_: (0,) * A.ndim),
		pl.BlockSpec(B.shape, lambda *_: (0,) * B.ndim),
	]

	out_specs = pl.BlockSpec((m, n), lambda *_: (0,) * A.ndim)
	return pl.pallas_call(
		f=partial(
			_gpu_matmul_kernel_fwd,
			blocksize_m=blocksize_m,
			blocksize_n=blocksize_n,
			blocksize_k=blocksize_k,
			prod_dtype=prod_dtype,
			precision=precision,
		),
		out_shape=jax.ShapeDtypeStruct(shape=(m, n), dtype=A.dtype),
		debug=False,
		interpret=INTERPRET,
		grid=grid,
		in_specs=in_specs,
		out_specs=out_specs,
		compiler_params=_get_compiler_params(
			blocksize_m=blocksize_m,
			blocksize_n=blocksize_n,
			blocksize_k=blocksize_k,
			dtype=A.dtype,
		),
		name="gpu_matmul_kernel_fwd",
	)(A, B)


def _call_gpu_matmul_kernel_fwd_residual(
	A: jax.Array,
	B: jax.Array,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	prod_dtype: jnp.dtype,
	precision: jax.lax.PrecisionLike = None,
):
	return _call_gpu_matmul_kernel_fwd(
		A=A,
		B=B,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		prod_dtype=prod_dtype,
		precision=precision,
	), (
		A,
		B,
	)


def _call_gpu_matmul_kernel_bwd(
	blocksize_m,
	blocksize_k,
	blocksize_n,
	prod_dtype,
	precision,
	res,
	gO,
):
	# A(mk)@B(kn)=C(mn)

	(
		A,
		B,
	) = res

	gA = _call_gpu_matmul_kernel_fwd(
		A=gO,
		B=B.transpose(1, 0),
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		prod_dtype=prod_dtype,
		precision=precision,
	)
	gB = _call_gpu_matmul_kernel_fwd(
		A=A.transpose(1, 0),
		B=gO,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		prod_dtype=prod_dtype,
		precision=precision,
	)
	return gA, gB


@partial(jax.custom_vjp, nondiff_argnums=(2, 3, 4, 5, 6))
def gpu_matmul(
	A: jax.Array,
	B: jax.Array,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	prod_dtype: jnp.dtype,
	precision: jax.lax.PrecisionLike = None,
):
	return _call_gpu_matmul_kernel_fwd(
		A=A,
		B=B,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		prod_dtype=prod_dtype,
		precision=precision,
	)


gpu_matmul.defvjp(_call_gpu_matmul_kernel_fwd_residual, _call_gpu_matmul_kernel_bwd)

__all__ = ["gpu_matmul"]



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for common sparse training algorithms."""

import functools

import chex
import jax
import jax.numpy as jnp
import optax
from absl.testing import absltest, parameterized

from fjformer.jaxpruner import sparsity_distributions
from fjformer.jaxpruner.algorithms import sparse_trainers


class SparseTrainerTest(parameterized.TestCase, absltest.TestCase):
	def testStaticRandomSparseDefaultArguments(self):
		self.assertTrue(sparse_trainers.StaticRandomSparse().is_sparse_gradients)

	def testSETDefaultArguments(self):
		updater = sparse_trainers.SET()
		self.assertTrue(updater.is_sparse_gradients)
		self.assertTrue(updater.skip_gradients)
		self.assertFalse(updater.is_debug)

	def testStaticRandomSparse(self):
		sparsity = 0.75
		sparsity_fn = functools.partial(sparsity_distributions.uniform, sparsity=sparsity)
		updater = sparse_trainers.StaticRandomSparse(sparsity_distribution_fn=sparsity_fn)
		param_tree = {"w": jnp.array([0.1, 0.9, 0.2, 0.5])}
		sparsity_tree = {"w": sparsity}
		initial_masks = updater.get_initial_masks(param_tree, sparsity_tree)
		self.assertEqual(jnp.sum(initial_masks["w"]), 1)

	@parameterized.parameters(
		# Masked drop [X, 1, X ,-1, X, 8, -2, X]
		# grow [9, 1, 4 ,1, 5, 8, 2, 3]
		[
			(0.0, [0, 1, 0, 1, 0, 1, 1, 0]),
			(0.25, [1, 1, 0, 1, 0, 1, 0, 0]),
			(0.5, [1, 1, 0, 0, 1, 1, 0, 0]),
			(0.75, [1, 0, 1, 0, 1, 1, 0, 0]),
			(1.0, [1, 0, 1, 0, 1, 1, 0, 0]),
		]
	)
	def testRigLUpdateMask(self, drop_fraction, expected_mask):
		expected_mask = jnp.array(expected_mask, dtype=jnp.uint8)
		sparsity_fn = functools.partial(sparsity_distributions.uniform, sparsity=0.5)
		updater = sparse_trainers.RigL(sparsity_distribution_fn=sparsity_fn, is_debug=True)
		old_mask = jnp.array([0, 1, 0, 1, 0, 1, 1, 0], dtype=jnp.uint8)
		param = jnp.array([-9, 1, 4, -1, 5, 8, -2, 3], dtype=jnp.float32)

		drop_score = param
		grow_score = jnp.abs(param)
		updated_mask = updater._update_masks(
			old_mask, drop_score, grow_score, drop_fraction=drop_fraction
		)
		self.assertSequenceAlmostEqual(expected_mask, updated_mask)

	@parameterized.product(
		updater=[sparse_trainers.RigL, sparse_trainers.SET],
		sparsity=[0.0, 0.5, 0.95],
		shape=[(1,), (5,), (4, 5), (2, 9, 1), (2, 2, 4, 4)],
	)
	def testDynamicSparseUpdateStateConstantSparsity(self, updater, sparsity, shape):
		dist_fn = functools.partial(
			sparsity_distributions.uniform,
			filter_fn=lambda *_: True,
			sparsity=sparsity,
		)
		updater = updater(sparsity_distribution_fn=dist_fn, skip_gradients=True)
		# Ensure weight diverge from zero
		params = {"w": jax.random.normal(jax.random.PRNGKey(0), shape)}
		optimizer = updater.wrap_optax(optax.sgd(0.1))
		opt_state = optimizer.init(params)
		grad = {"w": jax.random.normal(jax.random.PRNGKey(8), shape)}
		old_masks = opt_state.masks
		opt_state = updater.update_state(opt_state, params, grad)
		# Check masks have same sparsity.
		self.assertEqual(jnp.sum(old_masks["w"]), jnp.sum(opt_state.masks["w"]))

	@parameterized.product(
		shape=[(5,), (4, 5), (2, 9, 1), (2, 2, 4, 4)], is_adam=[False, True]
	)
	def testDynamicSparseUpdateStateInnerReset(self, shape, is_adam):
		dist_fn = functools.partial(
			sparsity_distributions.uniform, filter_fn=lambda *_: True, sparsity=0.5
		)
		updater = sparse_trainers.RigL(sparsity_distribution_fn=dist_fn)
		# Ensure weight diverge from zero
		params = {"w": jax.random.normal(jax.random.PRNGKey(0), shape)}
		if is_adam:
			optimizer = updater.wrap_optax(optax.adam(0.001))
			acc_names = ["mu", "nu"]
		else:
			optimizer = updater.wrap_optax(optax.sgd(0.001, momentum=0.9))
			acc_names = ["trace"]
		opt_state = optimizer.init(params)
		# Set trace to ones.
		acc_state = opt_state.inner_state[0]
		new_values = {
			k: jax.tree_map(jnp.ones_like, getattr(acc_state, k)) for k in acc_names
		}
		inner_state = (acc_state._replace(**new_values), *opt_state.inner_state[1:])
		opt_state = opt_state._replace(inner_state=inner_state)

		grad = {"w": jax.random.normal(jax.random.PRNGKey(8), shape)}
		old_masks = opt_state.masks
		opt_state = updater.update_state(opt_state, params, grad)
		# Check new connections have 0 trace
		new_connections = (old_masks["w"] == 0) & (opt_state.masks["w"] == 1)
		print(new_connections)
		for acc_key in acc_names:
			new_acc = getattr(opt_state.inner_state[0], acc_key)["w"]
			self.assertEqual(jnp.sum(new_acc[new_connections]), 0.0)
			self.assertNotEqual(jnp.sum(new_acc[~new_connections]), 0.0)

	def testRestartUsingMask(self):
		target_tree = {"a": [-9, 1, 4, -1, 5, 8, -2, 3]}
		masks = {"a": [0, 1, 0, 1, 0, 1, 1, 0]}
		expected = {"a": [-9, 0, 4, 0, 5, 0, 0, 3]}
		result = sparse_trainers._restart_using_mask(target_tree, masks)
		chex.assert_trees_all_close(result, expected)


if __name__ == "__main__":
	absltest.main()



# Copyright 2023 The EASYDEL Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation based on FlashAttention 2 (https://arxiv.org/pdf/2307.08691) by @erfanzar,
# with a few bug fixes and adjustments.

import functools
import math
from typing import Optional

import flax
import flax.linen
import jax
import jax.numpy as jnp
import jax.random as jrand
import jax.sharding
from fjformer.sharding import with_sharding_constraint
from fjformer import GenerateRNG
from jax import lax

rng = GenerateRNG()


@functools.partial(
	jax.jit,
	static_argnames=[
		"dtype",
		"precision",
		"q_block",
		"k_block",
	],
)
def flash_attention2(
	query_state: jax.Array,
	key_state: jax.Array,
	value_state: jax.Array,
	mask: Optional[jax.Array] = None,
	bias: Optional[jax.Array] = None,
	*,
	dropout: float = 0.0,
	inference: bool = True,
	key: Optional[jax.random.PRNGKey] = None,
	q_block: Optional[int] = None,
	k_block: Optional[int] = None,
	dtype: Optional[jnp.dtype] = None,
	precision: lax.PrecisionLike = None,
	head_dim: Optional[int] = None,
	softmax_scale: Optional[float] = None,
) -> jax.Array:
	"""
	Computes multi-head attention using FlashAttention implementation.

	This implementation makes use of the FlashAttention algorithm for faster
	and more memory-efficient computation of attention. It is particularly
	beneficial for long sequences.

	Args:
	  query_state: Query, shape (`batch_size`, `num_heads`, `q_len`, `head_dim`).
	  key_state: Key, shape (`batch_size`, `num_heads`, `kv_len`, `head_dim`).
	  value_state: Value, shape (`batch_size`, `num_heads`, `kv_len`, `head_dim`).
	  mask: Optional attention mask. This can be any of the following:

	    - No mask (default):  All attention weights are computed.
	    - Boolean mask (2D): shape (`batch_size`, `q_len`), with `True` for
	      valid and `False` for masked positions.
	    - Integer mask (2D): shape (`batch_size`, `q_len`), where the value at
	      each position indicates the length of the sequence to attend to.
	    - 3D mask: shape (`batch_size`, `q_len`, `kv_len`), with `True` for
	      valid and `False` for masked positions.

	  bias: Optional attention bias.
	  dropout: Dropout rate.
	  inference: Whether to run in inference mode.
	  key: PRNG key for dropout.
	  q_block: Block size for query processing.
	  k_block: Block size for key/value processing.
	  dtype: Optional dtype for the output.
	  precision: Optional precision for matrix multiplication.
		head_dim: Optional head dim to be used at `query_state = query_state / math.sqrt(float(head_dim or query_state.shape[-1]))`.
		softmax_scale Optional softmax_scale to be used for `query_state = query_state * softmax_scale`

	Returns:
	  Output of multi-head attention, with shape
	  (`batch_size`, `num_heads`, `q_len`, `head_dim`).

	Raises:
	  ValueError: If `dropout` is not in the range [0, 1], or if `key` is not
	    provided during training when `dropout` > 0.
	"""
	if not inference and dropout > 0 and key is None:
		raise ValueError("key must be provided for training")
	if dropout < 0 or dropout > 1:
		raise ValueError(f"invalid dropout {dropout}")
	if dtype is not None:
		query_state = query_state.astype(dtype)
		key_state = key_state.astype(dtype)

	k_block = min(key_state.shape[2], k_block or 128)
	q_block = min(query_state.shape[2], q_block or 128)
	if head_dim is not None and softmax_scale is not None:
		raise ValueError("you can't pass both `head_dim` and `softmax_scale`.")
	if head_dim is not None:
		query_state = query_state / math.sqrt(float(head_dim))
	elif softmax_scale is not None:
		query_state = query_state * softmax_scale
	else:
		query_state = query_state / math.sqrt(float(query_state.shape[-1]))
	return _flash_attn2(
		query_state,
		key_state,
		value_state,
		mask,
		bias,
		dropout,
		inference,
		key,
		q_block,
		k_block,
		dtype,
		precision,
	)


@functools.partial(
	jax.custom_vjp,
	nondiff_argnums=(
		5,
		6,
		7,
		8,
		9,
		10,
		11,
	),
)
def _flash_attn2(
	query_state: jax.Array,
	key_state: jax.Array,
	value_state: jax.Array,
	mask: Optional[jax.Array],
	bias: Optional[jax.Array],
	dropout: float,
	inference: bool,
	key: Optional[jax.random.PRNGKey],
	q_block: int,
	k_block: int,
	dtype: Optional[jnp.dtype],
	precision: lax.PrecisionLike,
) -> jax.Array:
	"""Custom VJP-enabled wrapper for FlashAttention forward pass."""
	return _fwd_flash_attn(
		query_state,
		key_state,
		value_state,
		mask,
		bias,
		dropout,
		inference,
		key,
		q_block,
		k_block,
		dtype,
		precision,
	)[0]


@functools.partial(jax.named_call, name="_fwd_flash_attn")
def _fwd_flash_attn(
	query_state: jax.Array,
	key_state: jax.Array,
	value_state: jax.Array,
	mask: Optional[jax.Array],
	bias: Optional[jax.Array],
	dropout: float,
	inference: bool,
	key: Optional[jax.random.PRNGKey],
	q_block: int,
	k_block: int,
	dtype: Optional[jnp.dtype],
	precision: lax.PrecisionLike,
) -> tuple[jax.Array, tuple[jax.Array, ...]]:
	"""Forward pass of FlashAttention."""
	b, h, _, d = query_state.shape
	q_seq = query_state.shape[2]
	k_seq = key_state.shape[2]
	assert q_seq % q_block == 0
	assert k_seq % k_block == 0
	Tr = q_seq // q_block
	Tc = k_seq // k_block
	o_shape = jax.eval_shape(
		lambda: (query_state @ key_state.transpose(0, 1, 3, 2)) @ value_state
	).shape
	o = jnp.zeros(o_shape, dtype=dtype)

	lse = jnp.full((b, h, q_seq), fill_value=-jnp.inf, dtype=jnp.float32)
	if hasattr(query_state, "sharding"):
		if isinstance(query_state.sharding, jax.sharding.NamedSharding):
			with query_state.sharding.mesh:
				o = with_sharding_constraint(o, query_state.sharding.spec)
				lse = with_sharding_constraint(
					lse,
					jax.sharding.PartitionSpec(*query_state.sharding.spec[:3]),
				)
		elif isinstance(
			query_state.sharding, jax.sharding.SingleDeviceSharding
		) and hasattr(query_state.sharding, "_device"):
			o = jax.device_put(o, query_state.sharding._device)
			lse = jax.device_put(lse, query_state.sharding._device)

	global_mask = mask

	@jax.jit
	@functools.partial(jax.named_call, name="_fwd_flash_attn_call_o")
	def call_o(state):
		i, o, lse = state
		q_i = jax.lax.dynamic_slice_in_dim(query_state, i * q_block, q_block, 2)
		o_i = jax.lax.dynamic_slice_in_dim(o, i * q_block, q_block, 2)
		lse_i = jax.lax.dynamic_slice_in_dim(lse, i * q_block, q_block, 2)
		m_i = jnp.full((b, h, q_block), fill_value=-jnp.inf, dtype=dtype)

		@jax.jit
		@functools.partial(jax.named_call, name="_fwd_flash_attn_call_o_call_qk")
		def call_qk(state):
			i, j, o_i, q_i, lse_i, m_i = state
			k_j = jax.lax.dynamic_slice_in_dim(key_state, j * k_block, k_block, 2)
			v_j = jax.lax.dynamic_slice_in_dim(value_state, j * k_block, k_block, 2)

			s_ij = jnp.einsum(
				"bhqd,bhdk->bhqk",
				q_i,
				k_j.transpose(0, 1, 3, 2),
				precision=precision,
			)

			if bias is not None:
				b_i = jax.lax.dynamic_slice_in_dim(bias, i * q_block, q_block, 2)
				b_ij = jax.lax.dynamic_slice_in_dim(b_i, j * k_block, k_block, 3)
				s_ij = s_ij + b_ij
			if global_mask is not None:
				ma_i = jax.lax.dynamic_slice_in_dim(
					global_mask, i * q_block, q_block, 2
				)
				ma_ij = jax.lax.dynamic_slice_in_dim(ma_i, j * k_block, k_block, 3)
				s_ij = jnp.where(ma_ij, s_ij, -1e10)

			if dropout > 0 and not inference:
				rng = jax.random.fold_in(key, i * Tc + j)
				keep_prob = 1.0 - dropout
				broadcast_shape = list(s_ij.shape)
				mask = jax.random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
				mask = jnp.broadcast_to(mask, s_ij.shape)
				s_ij = lax.select(mask, s_ij / keep_prob, jnp.zeros_like(s_ij))

			m_ij = jnp.maximum(m_i, jnp.max(s_ij, axis=-1))
			p = jnp.exp(s_ij - jnp.expand_dims(m_ij, -1))

			l_ij = jnp.sum(p, -1)

			o_scale = jnp.exp(m_i - m_ij)
			o_i = o_i * jnp.expand_dims(o_scale, -1)

			o_i = o_i + jnp.einsum(
				"bhqk,bhkd->bhqd",
				p,
				v_j,
				precision=precision,
			)

			return (
				i,
				j + 1,
				o_i.astype(dtype),
				q_i.astype(dtype),
				jnp.log(jnp.exp(lse_i - m_ij) + l_ij) + m_ij,
				m_ij.astype(dtype),
			)

		j_end = jnp.minimum(i + 1, Tc) if mask is not None else Tc

		_, _, o_i, _, lse_i, m_i = jax.lax.while_loop(
			lambda state: state[1] < j_end,
			call_qk,
			(i, 0, o_i, q_i, lse_i, m_i),
		)
		o_scale = jnp.exp(m_i - lse_i)
		o_i = o_i * jnp.expand_dims(o_scale, -1)

		o = jax.lax.dynamic_update_slice_in_dim(o, o_i.astype(o.dtype), i * q_block, 2)
		lse = jax.lax.dynamic_update_slice_in_dim(
			lse,
			lse_i.astype(lse.dtype),
			i * q_block,
			2,
		)
		return i + 1, o, lse

	_, o, lse = jax.lax.while_loop(lambda state: state[0] < Tr, call_o, (0, o, lse))

	return o, (
		o,
		lse,
		query_state,  #: jax.Array
		key_state,  #: jax.Array
		value_state,  #: jax.Array
		mask,
		bias,
	)


@functools.partial(jax.named_call, name="_bwd_flash_attn")
def _bwd_flash_attn(
	dropout: float,
	inference: bool,
	key: Optional[jax.random.PRNGKey],
	q_block: int,
	k_block: int,
	dtype: Optional[jnp.dtype],
	precision: lax.PrecisionLike,
	residuals,
	grad_in: jax.Array,
):
	"""Backward pass of FlashAttention."""

	del dtype
	(
		O,  # noqa: E741
		L,
		query_state,
		key_state,
		value_state,
		mask,
		bias,
	) = residuals
	dO = grad_in

	b, h, _, d = query_state.shape
	q_seq = query_state.shape[2]
	k_seq = key_state.shape[2]
	assert q_seq % q_block == 0
	assert k_seq % k_block == 0
	Tr = q_seq // q_block
	Tc = k_seq // k_block

	D = jnp.sum(dO * O, axis=-1)

	dQ = (query_state * 0.0).astype(query_state.dtype)
	dK = (key_state * 0.0).astype(key_state.dtype)
	dV = (value_state * 0.0).astype(value_state.dtype)
	global_mask = mask
	is_causal = mask is not None

	@jax.jit
	@functools.partial(jax.named_call, name="_bwd_flash_attn_call_o")
	def call_o(state):
		j, dQ, dK, dV = state
		k_j = jax.lax.dynamic_slice_in_dim(key_state, j * k_block, k_block, 2)
		v_j = jax.lax.dynamic_slice_in_dim(value_state, j * k_block, k_block, 2)

		dK_j = jax.lax.dynamic_slice_in_dim(dK, j * k_block, k_block, 2)
		dV_j = jax.lax.dynamic_slice_in_dim(dV, j * k_block, k_block, 2)

		@jax.jit
		@functools.partial(jax.named_call, name="_bwd_flash_attn_call_o_call_qk")
		def do_inner_block(state):
			i, j, dQ, dK_j, dV_j = state
			q_i = jax.lax.dynamic_slice_in_dim(query_state, i * q_block, q_block, 2)
			dQ_i = jax.lax.dynamic_slice_in_dim(dQ, i * q_block, q_block, 2)
			dO_i = jax.lax.dynamic_slice_in_dim(dO, i * q_block, q_block, 2)

			L_i = jax.lax.dynamic_slice_in_dim(L, i * q_block, q_block, 2)
			D_i = jax.lax.dynamic_slice_in_dim(D, i * q_block, q_block, 2)
			s_ij = q_i @ k_j.transpose(0, 1, 3, 2)
			if dropout > 0 and not inference:
				rng = jax.random.fold_in(key, i * Tc + j)
				keep_prob = 1.0 - dropout
				broadcast_shape = list(s_ij.shape)
				mask = jax.random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
				mask = jnp.broadcast_to(mask, s_ij.shape)
				s_ij = lax.select(mask, s_ij / keep_prob, jnp.zeros_like(s_ij))

			if bias is not None:
				b_i = jax.lax.dynamic_slice_in_dim(bias, i * q_block, q_block, 2)
				b_ij = jax.lax.dynamic_slice_in_dim(b_i, j * k_block, k_block, 3)
				s_ij = s_ij + b_ij

			if global_mask is not None:
				ma_i = jax.lax.dynamic_slice_in_dim(
					global_mask, i * q_block, q_block, 2
				)
				ma_ij = jax.lax.dynamic_slice_in_dim(ma_i, j * k_block, k_block, 3)
				s_ij = jnp.where(ma_ij, s_ij, -1e10)

			p_ij = jnp.exp(s_ij - jnp.expand_dims(L_i, -1))

			if dropout > 0 and not inference:
				rng = jax.random.fold_in(key, i * Tc + j)
				keep_prob = 1.0 - dropout
				broadcast_shape = list(p_ij.shape)
				mask = jax.random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
				mask = jnp.broadcast_to(mask, p_ij.shape)
				p_ij = lax.select(mask, p_ij / keep_prob, jnp.zeros_like(p_ij))

			dV_j = dV_j + jnp.matmul(
				p_ij.transpose(0, 1, 3, 2), dO_i, precision=precision
			)

			dP_ij = jnp.matmul(dO_i, v_j.transpose(0, 1, 3, 2), precision=precision)

			dS_ij = p_ij * (dP_ij - D_i[..., None])
			dQ_i = dQ_i + jnp.matmul(dS_ij, k_j, precision=precision)
			dK_j = dK_j + jnp.matmul(
				dS_ij.transpose(0, 1, 3, 2), q_i, precision=precision
			)
			dQ = jax.lax.dynamic_update_slice_in_dim(
				dQ,
				dQ_i.astype(dQ.dtype),
				i * q_block,
				2,
			)
			return (
				i + 1,
				j,
				dQ.astype(query_state.dtype),
				dK_j.astype(key_state.dtype),
				dV_j.astype(value_state.dtype),
			)

		i_start = j if is_causal else 0
		_, j, dQ, dK_j, dV_j = jax.lax.while_loop(
			lambda state: state[0] < Tr,
			do_inner_block,
			(i_start, j, dQ, dK_j, dV_j),
		)

		dK = jax.lax.dynamic_update_slice_in_dim(
			dK, dK_j.astype(dK.dtype), j * q_block, 2
		)
		dV = jax.lax.dynamic_update_slice_in_dim(
			dV, dV_j.astype(dV.dtype), j * q_block, 2
		)

		return j + 1, dQ, dK, dV

	_, dQ, dK, dV = jax.lax.while_loop(
		lambda state: state[0] < Tc,
		call_o,
		(0, dQ, dK, dV),
	)

	return dQ, dK, dV, None, None


_flash_attn2.defvjp(_fwd_flash_attn, _bwd_flash_attn)


def fwd_test():
	b, h, qs, s, d = 1, 8, 32, 128, 128
	dtype = jnp.float16

	q = jrand.normal(rng.rng, shape=(b, qs, h, d), dtype=dtype)
	k = jrand.normal(rng.rng, shape=(b, s, h, d), dtype=dtype)
	v = jrand.normal(rng.rng, shape=(b, s, h, d), dtype=dtype)
	b = jnp.where(
		jrand.randint(rng.rng, shape=(b, h, qs, s), minval=0, maxval=3) > 1,
		0,
		jnp.finfo(dtype).min,
	)

	excepted_result = flax.linen.attention.dot_product_attention(
		query=q,
		key=k,
		value=v,
		bias=b,
	)
	result = flash_attention2(
		query_state=q.transpose(0, 2, 1, 3),
		key_state=k.transpose(0, 2, 1, 3),
		value_state=v.transpose(0, 2, 1, 3),
		bias=b,
		dtype=dtype,
		q_block=64,
		k_block=64,
	).transpose(0, 2, 1, 3)

	print(f"PRED : {result[0,0,0,:5]}")
	print(f"ORGN : {excepted_result[0,0,0,:5]}")

	print(jnp.allclose(excepted_result, result, atol=0.125, rtol=0))


def bwd_test():
	b, h, qs, s, d = 2, 32, 64, 64, 64
	dtype = jnp.float32

	q = jrand.normal(rng.rng, shape=(b, qs, h, d), dtype=dtype)
	k = jrand.normal(rng.rng, shape=(b, s, h, d), dtype=dtype)
	v = jrand.normal(rng.rng, shape=(b, s, h, d), dtype=dtype)
	b = jnp.where(
		jrand.randint(rng.rng, shape=(b, h, qs, s), minval=0, maxval=3) > 1,
		0,
		jnp.finfo(dtype).min,
	)

	excepted_result = jax.grad(
		lambda *x: flax.linen.attention.dot_product_attention(*x).sum()
	)(q, k, v)
	result = jax.grad(
		lambda *x: flash_attention2(
			*x,
			dtype=dtype,
			q_block=qs,
			k_block=s,
			precision=jax.lax.Precision("HIGHEST".lower()),
		).sum()
	)(
		q.transpose(0, 2, 1, 3),
		k.transpose(0, 2, 1, 3),
		v.transpose(0, 2, 1, 3),
	).transpose(0, 2, 1, 3)

	print(f"PRED BWD : {result[0,0,0,:5]}")
	print(f"ORGN BWD : {excepted_result[0,0,0,:5]}")

	print(jnp.allclose(excepted_result, result, atol=0.125, rtol=0))


jax_flash_attn_2_mu = _flash_attn2

__all__ = ["jax_flash_attn_2_mu"]

if __name__ == "__main__":
	fwd_test()
	bwd_test()



"""Utilities for memory profiling in JAX."""

import curses
import json
import os
import re
import subprocess
import threading
import time
from typing import Any, Dict, Optional

import IPython.display
import jax
import jax.profiler

try:
	import posix
except ModuleNotFoundError:
	posix = None


def is_notebook() -> bool:
	"""Returns True if the code is being run in a notebook, False otherwise."""
	return os.environ.get("IPYTHON") is not None


def run(
	note_book: Optional[bool] = None,
	interval: float = 1,
	dir_prefix: str = "/dev/shm",
	display_output: bool = True,
) -> None:
	"""Periodically prints JAX memory usage information.

	Runs the `go tool pprof` command in a loop, capturing its output and either
	printing it to stdout or displaying it in a Jupyter notebook cell.

	Args:
	    note_book: Whether the code is running in a notebook. If None, it will
	        be auto-detected.
	    interval: The time interval (in seconds) between refreshes.
	    dir_prefix: The directory to store the memory profile.
	    display_output: Whether to print the output to the console or a
	        Jupyter notebook cell.
	"""
	if note_book is None:
		note_book = is_notebook()

	std = curses.initscr() if not note_book else None
	try:
		while True:
			if not note_book and display_output:
				std.clear()
			output = subprocess.run(
				args=["go", "tool", "pprof", "-tags", f"{dir_prefix}/memory.prof"],
				stdout=subprocess.PIPE,
				stderr=subprocess.DEVNULL,
			).stdout.decode("utf-8")

			if display_output:
				if not note_book:
					std.addstr(output)
					std.refresh()
				else:
					IPython.display.clear_output(True)
					print(output)

			with open(f"{dir_prefix}/memory.json", "w") as fin:
				json.dump({"log": output}, fin)
			time.sleep(interval)
	except KeyboardInterrupt:
		curses.endwin()


def get_memory_information(dir_prefix: str = "/dev/shm") -> str:
	"""Retrieves JAX memory usage information using `go tool pprof`.

	Args:
	    dir_prefix: The directory where the memory profile is stored.

	Returns:
	    The output of the `go tool pprof` command as a string.
	"""
	return subprocess.run(
		args=["go", "tool", "pprof", "-tags", f"{dir_prefix}/memory.prof"],
		stdout=subprocess.PIPE,
		stderr=subprocess.DEVNULL,
	).stdout.decode("utf-8")


def initialise_tracking(interval: float = 1.0, dir_prefix: str = "/dev/shm") -> None:
	"""Starts a daemon thread to periodically save the memory profile to disk.

	This function starts a background thread that continuously saves the JAX
	memory profile to a file. It ensures atomicity of the file update using
	`posix.rename` if available, otherwise falls back to a simple rename.

	Args:
	    interval: The time interval (in seconds) between memory profile saves.
	    dir_prefix: The directory to store the memory profile.
	"""

	def _save_profile():
		while True:
			jax.profiler.save_device_memory_profile(f"{dir_prefix}/memory.prof.new")
			if posix is not None:
				os.rename(f"{dir_prefix}/memory.prof.new", f"{dir_prefix}/memory.prof")
			else:
				posix.rename(f"{dir_prefix}/memory.prof.new", f"{dir_prefix}/memory.prof")
			time.sleep(interval)

	thread = threading.Thread(target=_save_profile, daemon=True)
	thread.start()


def threaded_log(
	interval: float = 1.0, dir_prefix: str = "/dev/shm", save_mem_json: bool = False
) -> threading.Thread:
	"""Starts a thread to periodically log memory information.

	This function launches a background thread that continuously monitors and
	either prints or displays the JAX memory usage information. Optionally,
	it can also save the memory information to a JSON file.

	Args:
	    interval: The time interval (in seconds) between memory logs.
	    dir_prefix: The directory to save the memory information JSON file.
	    save_mem_json: Whether to save the memory information to a JSON file.

	Returns:
	    The thread handling the memory logging.
	"""
	note_book = is_notebook()

	def _show_memory_info():
		std = curses.initscr() if not note_book else None
		try:
			while True:
				mem_info = get_memory_information()
				if not note_book:
					std.clear()
					std.addstr(mem_info)
					std.refresh()
				else:
					IPython.display.clear_output(True)
					print(mem_info)
				if save_mem_json:
					with open(f"{dir_prefix}/memory.json", "w") as fin:
						json.dump({"log": mem_info}, fin)
				time.sleep(interval)
		except KeyboardInterrupt:
			curses.endwin()

	thread = threading.Thread(target=_show_memory_info)
	return thread


def get_capacity_matrix(dir_prefix: str = "/dev/shm") -> Dict[str, Dict[str, Any]]:
	"""Parses memory information and returns a dictionary with capacity details.

	This function extracts memory usage information from the output of the
	`go tool pprof` command and structures it into a dictionary. The dictionary
	maps memory pool names to their usage details, including used memory,
	usage percentage, process information, and full capacity.

	Args:
	    dir_prefix: The directory where the memory profile is stored.

	Returns:
	    A dictionary containing memory capacity information.
	"""
	pattern = r"(\d+\.\d+\wB) \((\d+\.\d+%)\): (\w+)(\(.*?\))?"

	def _calculate_full_size(size: str, percent: str) -> float:
		size_in_gb = float(re.search(r"(\d+\.\d+)GB", size).group(1))
		percent_value = 100 / float(re.search(r"(\d+\.\d+)%", percent).group(1))
		return size_in_gb * percent_value

	matches = re.findall(pattern, get_memory_information(dir_prefix=dir_prefix))
	information = {}
	try:
		for match in matches:
			information[match[2]] = {
				"Used": match[0],
				"Usage Percent": match[1],
				"Process": match[3][1:] if match[3] else "∞",
				"Full Capacity": _calculate_full_size(match[0], match[1]),
			}
	except (ArithmeticError, AttributeError, KeyError, ValueError):
		pass
	return information



# TODO : Implement Custom Backward Prp
from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import Any, Optional, Sequence, Union

import jax
from jax import Array, lax
from jax import numpy as jnp
from jax.core import Primitive

import fjformer.core as core


def create_dynamic_map(signed=True, max_exponent_bits=7, total_bits=8):
	"""
	Creates the dynamic quantization map.
	"""

	data = []
	non_sign_bits = total_bits - (1 if signed else 0)
	additional_items = 2 ** (non_sign_bits - max_exponent_bits) - 1
	for i in range(max_exponent_bits):
		fraction_items = int(
			(
				2 ** (i + non_sign_bits - max_exponent_bits) + 1
				if signed
				else 2 ** (i + non_sign_bits - max_exponent_bits + 1) + 1
			),
		)
		boundaries = jnp.linspace(0.1, 1, fraction_items)
		means = (boundaries[:-1] + boundaries[1:]) / 2.0
		data += ((10 ** (-(max_exponent_bits - 1) + i)) * means).tolist()
		if signed:
			data += (-(10 ** (-(max_exponent_bits - 1) + i)) * means).tolist()

	if additional_items > 0:
		boundaries = jnp.linspace(0.1, 1, additional_items + 1)
		means = (boundaries[:-1] + boundaries[1:]) / 2.0
		data += (
			(10 ** (-(max_exponent_bits - 1) + (max_exponent_bits - 1))) * means
		).tolist()
		if signed:
			data += (
				-(10 ** (-(max_exponent_bits - 1) + (max_exponent_bits - 1))) * means
			).tolist()

	data.append(0)
	data.append(1.0)

	assert len(data) == 2**total_bits

	gap = 256 - len(data)
	for _ in range(gap):
		data.append(0)

	data.sort()
	return jnp.array(data, dtype=jnp.float32)


CODE = create_dynamic_map()
CODE = CODE.at[0].set(-1.0)
OLD_QUANT = True
DEFAULT_BLOCKSIZE = 1024


@jax.jit
def quantize_block_jax(block):
	# 1. find absmax in block
	absmax_block = jnp.max(jnp.abs(block))

	# 2. divide input value by absmax to normalize into [-1.0, 1.0]
	normed_values = block / absmax_block

	# 3. do binary search to find the closest value
	idx = jnp.searchsorted(CODE, normed_values)

	# 4. check minimal distance
	def check_distance(idx, normed_value):
		dist_left = jnp.abs(normed_value - CODE[idx])
		dist_right = jnp.abs(normed_value - CODE[idx + 1])
		return jnp.where(
			idx < 255,
			jnp.where(dist_right < dist_left, idx + 1, idx),
			idx,
		)

	idx = jax.vmap(check_distance)(idx, normed_values)

	# 5. store index
	return idx.astype(jnp.uint8), absmax_block


@functools.partial(jax.jit, static_argnames=["blocksize"])
def quantize_jax(A, blocksize):
	# Ensure code[0] is -1.0

	# Flatten the input array
	A_flat = A.ravel()

	# Pad A_flat to be divisible by blocksize
	pad_size = (blocksize - A_flat.size % blocksize) % blocksize
	A_padded = jnp.pad(A_flat, (0, pad_size))

	# Reshape A_padded into blocks
	num_blocks = A_padded.size // blocksize
	A_blocks = A_padded.reshape(num_blocks, blocksize)

	# Apply quantize_block to each block
	quantized_blocks, absmax = jax.vmap(quantize_block_jax, in_axes=(0,))(A_blocks)

	# Flatten the quantized blocks and trim to original size
	quantized = quantized_blocks.ravel()[: A_flat.size]

	# Reshape quantized back to the original shape
	quantized = quantized.reshape(A.shape)

	return quantized, absmax.astype(jnp.float32)


if OLD_QUANT:

	@functools.partial(jax.jit, static_argnames=["blocksize"])
	def dequantize_jax(A, absmax, blocksize):
		"""
		Dequantize the input array.

		Args:
		    A (jnp.ndarray): The quantized array.
		    absmax (jnp.ndarray): The absolute maximum values for each block.
		    blocksize (int): The size of the blocks for dequantization.

		Returns:
		    jnp.ndarray: The dequantized array.
		"""
		indices = jnp.arange(A.shape[0])
		block_indices = indices // blocksize
		extra_dim = ()
		if CODE[A].ndim > 1:
			extra_dim = (1,) * (len(CODE[A].shape) - 1)
		dequantized_A = CODE[A] * jnp.reshape(
			absmax[block_indices],
			(CODE[A].shape[0],) + extra_dim,
		)
		return dequantized_A

else:

	@functools.partial(jax.jit, static_argnames=["blocksize"])
	def dequantize_jax(A, absmax, blocksize):
		# Flatten A
		A_flat = A.ravel()

		# Compute the number of blocks
		num_blocks = (A_flat.size + blocksize - 1) // blocksize

		# Pad A to be divisible by blocksize
		pad_size = (blocksize - A_flat.size % blocksize) % blocksize
		A_padded = jnp.pad(A_flat, (0, pad_size))

		# Reshape A_padded into blocks
		A_blocks = A_padded.reshape(num_blocks, blocksize)

		# Ensure absmax has the correct shape
		absmax = absmax[:num_blocks]

		# Define the dequantization function for a single block
		def dequantize_block(block, block_absmax):
			return CODE[block] * block_absmax

		# Apply dequantize_block to each block
		out_blocks = jax.vmap(dequantize_block)(A_blocks, absmax)

		# Flatten and trim to original size
		out = out_blocks.ravel()[: A_flat.size]

		# Reshape out to match the original shape of A
		out = out.reshape(A.shape)

		return out


@dataclass
class Array8Lt(core.ImplicitArray):
	"""
	Custom 8-bit Quantized Array for efficient manipulation of JAX arrays.

	This class provides methods for quantizing and dequantizing JAX arrays to 8-bit
	representation, which can significantly reduce memory usage and potentially
	improve computation speed for certain operations.

	Example:
	    >>> import jax
	    >>> import jax.numpy as jnp

	    >>> x = jax.random.normal(jax.random.key(0), (512, 64), dtype=jnp.float32)
	    >>> xp = jax.random.normal(jax.random.key(1), (64, 256), dtype=jnp.float32)

	    >>> quantized_x = Array8Lt.quantize(
	    ...   x,
	    ... )
	    >>> quantized_xp = Array8Lt.quantize(
	    ...   xp,
	    ... )

	    >>> @jax.jit
	    >>> @core.implicit_compact
	    >>> def f(a, b):
	    ...   return jnp.dot(a, b)

	    >>> result = f(x, xp)
	    >>> q_result = f(quantized_x, quantized_xp)

	    >>> print(jnp.allclose(result, q_result, rtol=1e-2, atol=1e-2))
	    True
	"""

	A: core.ArrayValue
	absmax: core.ArrayValue
	blocksize: int = core.aux_field()

	def materialize(self) -> Array:
		"""
		Materialize the quantized array back to its original representation.

		Returns:
		    Array: The dequantized array in its original dtype.
		"""
		return self.dequantize(
			A=self.A,
			absmax=self.absmax,
			blocksize=self.blocksize,
			float_dtype=self.dtype,
		)

	@classmethod
	def quantize(
		cls,
		array: Array,
		blocksize: int = DEFAULT_BLOCKSIZE,
		dtype: Optional[jnp.dtype] = None,
	) -> Array8Lt:
		QA, absmax = quantize_jax(array, blocksize)
		return cls(
			A=QA,
			absmax=absmax,
			shape=QA.shape,
			blocksize=blocksize,
			dtype=dtype or array.dtype,
		)

	@staticmethod
	def dequantize(
		A,
		absmax,
		blocksize,
		float_dtype: jnp.dtype,
	) -> Array:
		return dequantize_jax(
			A=A,
			absmax=absmax.astype(jnp.float32),
			blocksize=blocksize,
		).astype(float_dtype)

	def __repr__(self) -> str:
		return f"Array8Lt(shape={self.shape}, dtype={self.dtype})"

	@property
	def nbytes(self) -> int:
		"""
		Calculate the total number of bytes used by the quantized representation.

		Returns:
		    int: The number of bytes used.
		"""
		return self.A.nbytes + self.absmax.nbytes

	def memory_savings(self) -> float:
		"""
		Calculate the memory savings compared to the original array.

		Returns:
		    float: The percentage of memory saved.
		"""
		original_size = jnp.prod(jnp.array(self.shape)) * jnp.dtype(self.dtype).itemsize
		return (1 - self.nbytes / original_size) * 100


ArrayType = Union[Array, Array8Lt]


@core.primitive_handler("dot_general")
def handle_dot_general(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's dot_general operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array.
	    rhs (ArrayType): Right-hand side array.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.dot_general operation.
	"""
	if isinstance(lhs, Array8Lt):
		lhs = lhs.materialize()
	if isinstance(rhs, Array8Lt):
		rhs = rhs.materialize()
	return lax.dot_general(lhs, rhs, *args, **kwargs)


@core.primitive_handler("add")
def handle_add(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's add operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to add.
	    y (ArrayType): Second array to add.

	Returns:
	    The result of lax.add operation.
	"""
	if isinstance(x, Array8Lt):
		x = x.materialize()
	if isinstance(y, Array8Lt):
		y = y.materialize()
	return lax.add(x, y)


@core.primitive_handler("reduce")
def handle_reduce(
	primitive,
	operand: ArrayType,
	init_value: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's reduce operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be reduced.
	    init_value (ArrayType): The initial value for the reduction.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.reduce operation.
	"""
	if isinstance(operand, Array8Lt):
		operand = operand.materialize()
	if isinstance(init_value, Array8Lt):
		init_value = init_value.materialize()
	return lax.reduce(operand, init_value, *args, **kwargs)


@core.primitive_handler("mul")
def handle_mul(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's mul operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to multiply.
	    y (ArrayType): Second array to multiply.

	Returns:
	    The result of lax.mul operation.
	"""
	if isinstance(x, Array8Lt):
		x = x.materialize()
	if isinstance(y, Array8Lt):
		y = y.materialize()
	return lax.mul(x, y)


@core.primitive_handler("transpose")
def handle_transpose(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's transpose operation.

	Materializes Array8Lt input before performing the operation.
	Re-quantizes the result if the input was Array8Lt.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be transposed.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.transpose operation, potentially re-quantized.
	"""
	original_quantized = False
	if isinstance(operand, Array8Lt):
		blocksize = operand.blocksize
		operand = operand.materialize()
		original_quantized = True
	operand = lax.transpose(operand, *args, **kwargs)
	if original_quantized:
		operand = Array8Lt.quantize(
			operand,
			blocksize=blocksize,
			dtype=operand.dtype,
		)
	return operand


@core.primitive_handler("conv_general_dilated")
def handle_conv(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's conv_general_dilated operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array (input).
	    rhs (ArrayType): Right-hand side array (kernel).
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.conv operation.
	"""
	if isinstance(lhs, Array8Lt):
		lhs = lhs.materialize()
	if isinstance(rhs, Array8Lt):
		rhs = rhs.materialize()
	return lax.conv_general_dilated(lhs, rhs, *args, **kwargs)


@core.primitive_handler("max")
def handle_max(
	primitive,
	x: ArrayType,
	y: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's max operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array for max comparison.
	    y (ArrayType): Second array for max comparison.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.max operation.
	"""
	if isinstance(x, Array8Lt):
		x = x.materialize()
	if isinstance(y, Array8Lt):
		y = y.materialize()
	return lax.max(x, y, *args, **kwargs)


@core.primitive_handler("exp")
def handle_exp(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's exp operation.

	Materializes Array8Lt input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply exponential to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.exp operation.
	"""
	if isinstance(x, Array8Lt):
		x = x.materialize()
	return lax.exp(x, *args, **kwargs)


@core.primitive_handler("log")
def handle_log(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's log operation.

	Materializes Array8Lt input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply logarithm to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.log operation.
	"""
	if isinstance(x, Array8Lt):
		x = x.materialize()
	return lax.log(x, *args, **kwargs)


@core.primitive_handler("reshape")
def handle_reshape(
	primitive: Primitive,
	operand: ArrayType,
	**kwargs: Any,
):
	"""
	Custom handler for JAX's reshape operation.

	This function handles reshaping for both regular arrays and Array8Lt quantized arrays.
	It materializes Array8Bit input before reshaping and re-quantizes the result if the input was Array8Bit.

	Args:
	    primitive (Primitive): The JAX primitive being handled.
	    operand (ArrayType): The array to be reshaped.
	    new_sizes (Tuple[int, ...]): The desired new shape of the array.
	    dimensions (Tuple[int, ...], optional): The order in which dimensions should be permuted before reshaping.
	    **kwargs: Additional keyword arguments for the reshape operation.

	Returns:
	    ArrayType: The reshaped array, potentially re-quantized if the input was Array8Lt.

	Raises:
	    ValueError: If the new shape is not compatible with the original array's size.
	"""
	original_quantized = isinstance(operand, Array8Lt)

	if original_quantized:
		blocksize = operand.blocksize
		operand = operand.materialize()

	try:
		reshaped = lax.reshape(operand, **kwargs)
	except ValueError as e:
		raise ValueError(
			f"Reshape operation failed: {str(e)}. "
			f"Ensure the new shape {kwargs} is compatible with the original array size."
		) from e
	if original_quantized:
		reshaped = Array8Lt.quantize(
			operand,
			blocksize=blocksize,
			dtype=operand.dtype,
		)
	return reshaped


@core.primitive_handler("concatenate")
def handle_concatenate(
	primitive,
	operands: Sequence[ArrayType],
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's concatenate operation.

	Materializes Array8Lt inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operands (Sequence[ArrayType]): Sequence of arrays to concatenate.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.concatenate operation.
	"""
	materialized_operands = [
		op.materialize() if isinstance(op, Array8Lt) else op for op in operands
	]
	return lax.concatenate(materialized_operands, *args, **kwargs)


@core.primitive_handler("convert_element_type")
def convert_element_type(
	primitive,
	arg: ArrayType,
	**params,
) -> ArrayType:
	"""Handle element type conversion for Array8Lt."""
	result = jax.tree_util.tree_map(
		functools.partial(core.default_handler, primitive, **params), arg
	)
	result.dtype = params["new_dtype"]
	return result


@core.primitive_handler("broadcast_in_dim")
def handle_broadcast_in_dim(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle broadcast_in_dim for Array8Lt."""
	original_quantized = isinstance(operand, Array8Lt)
	array = operand
	if original_quantized:
		blocksize = operand.blocksize
		array = operand.materialize()
	result = jax.lax.broadcast_in_dim(array, *args, **kwargs)
	if original_quantized:
		result = Array8Lt.quantize(
			result,
			blocksize=blocksize,
			dtype=operand.dtype,
		)
	return result


@core.primitive_handler("gather")
def handle_gather(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle gather for Array8Lt."""
	original_quantized = isinstance(operand, Array8Lt)
	array = operand
	if original_quantized:
		array = operand.materialize()
	result = jax.lax.gather(array, *args, **kwargs)
	return result



from dataclasses import dataclass
from functools import partial
from typing import Any, Optional, Sequence, Union

import jax
from jax import Array, lax
from jax import numpy as jnp
from jax.core import Primitive

import fjformer.core as core


@jax.jit
def quantize_row_q4_0(array):
	scale = jnp.max(jnp.abs(array), axis=-1, keepdims=True)
	array = jnp.int4(jnp.rint(array * (7.5 / scale)))
	return array, scale


@jax.jit
def dequantize_row_q4_0(quants, scales):
	dequantized = quants.astype(jnp.float32) * (scales / 7.5)
	return dequantized


@dataclass
class A4Q(core.ImplicitArray):
	"""
	Custom 4-bit Quantized Array for efficient manipulation of JAX arrays.

	This class provides methods for quantizing and dequantizing JAX arrays to 4-bit
	representation, which can significantly reduce memory usage and potentially
	improve computation speed for certain operations.

	Attributes:
	    array_quantized (core.ArrayValue): The quantized array data.
	    scale (core.ArrayValue): Scaling factors for dequantization.
	    shape (tuple): Shape of the quantized array.
	    dtype (jnp.dtype): Original dtype of the array before quantization.

	Example:
	    >>> import jax
	    >>> import jax.numpy as jnp

	    >>> x = jax.random.normal(jax.random.key(0), (512, 64), dtype=jnp.float32)
	    >>> xp = jax.random.normal(jax.random.key(1), (64, 256), dtype=jnp.float32)

	    >>> quantized_x = A4Q.quantize(x)
	    >>> quantized_xp = A4Q.quantize(xp)

	    >>> @jax.jit
	    >>> @core.implicit_compact
	    >>> def f(a, b):
	    ...   return jnp.dot(a, b)

	    >>> result = f(x, xp)
	    >>> q_result = f(quantized_x, quantized_xp)

	    >>> print(jnp.allclose(result, q_result, rtol=1e-2, atol=1e-2))
	    True
	"""

	array_quantized: core.ArrayValue
	scale: core.ArrayValue

	def materialize(self) -> Array:
		"""
		Materialize the quantized array back to its original representation.

		Returns:
		    Array: The dequantized array in its original dtype.
		"""
		return self.dequantize(
			array_quantized=self.array_quantized,
			scale=self.scale,
			float_dtype=self.dtype,
			shape=self.shape,
		)

	@classmethod
	def quantize(
		cls,
		array: Array,
		dtype: Optional[jnp.dtype] = None,
		q4: int = 128,
		*_,
		**__,
	) -> "A4Q":
		"""
		Quantize a JAX array to 4-bit representation.

		Args:
		    array (Array): The input array to quantize.
		    dtype (jnp.dtype, optional): The desired dtype for the output. If None, uses the input array's dtype.

		Returns:
		    A4Q: The quantized array.
		"""

		org_shape = array.shape
		q4 = min(q4, array.size)
		if q4 % array.size != 0:
			q4 = array.shape[-1]
		array = array.reshape(-1, q4)
		quants, scales = quantize_row_q4_0(array=array)

		return cls(
			array_quantized=quants,
			scale=scales,
			shape=org_shape,
			dtype=dtype or array.dtype,
		)

	@staticmethod
	def dequantize(
		array_quantized: Array,
		scale: Array,
		float_dtype: jnp.dtype,
		shape: Sequence[int],
	) -> Array:
		"""
		Dequantize an 4-bit array back to its original representation.

		Args:
		    array_quantized (Array): The quantized array data.
		    scale (Array): The scaling factors used in quantization.
		    float_dtype (jnp.dtype): The desired output dtype.
				shape (Sequence[int]): org array shape
		Returns:
		    Array: The dequantized array.
		"""

		array = (
			dequantize_row_q4_0(
				quants=array_quantized,
				scales=scale,
			)
			.reshape(shape)
			.astype(float_dtype)
		)

		return array

	def __repr__(self) -> str:
		return f"A4Q(quants={self.array_quantized}, shape={self.shape}, dtype={self.dtype})"

	@property
	def nbytes(self) -> int:
		"""
		Calculate the total number of bytes used by the quantized representation.

		Returns:
		    int: The number of bytes used.
		"""
		return self.array_quantized.nbytes + self.scale.nbytes

	def memory_savings(self) -> float:
		"""
		Calculate the memory savings compared to the original array.

		Returns:
		    float: The percentage of memory saved.
		"""
		original_size = jnp.prod(jnp.array(self.shape)) * jnp.dtype(self.dtype).itemsize
		return (1 - self.nbytes / original_size) * 100


ArrayType = Union[Array, A4Q]


@core.primitive_handler("dot_general")
def handle_dot_general(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's dot_general operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array.
	    rhs (ArrayType): Right-hand side array.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.dot_general operation.
	"""
	if isinstance(lhs, A4Q):
		lhs = lhs.materialize()
	if isinstance(rhs, A4Q):
		rhs = rhs.materialize()
	return lax.dot_general(lhs, rhs, *args, **kwargs)


@core.primitive_handler("add")
def handle_add(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's add operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to add.
	    y (ArrayType): Second array to add.

	Returns:
	    The result of lax.add operation.
	"""
	if isinstance(x, A4Q):
		x = x.materialize()
	if isinstance(y, A4Q):
		y = y.materialize()
	return lax.add(x, y)


@core.primitive_handler("reduce")
def handle_reduce(
	primitive,
	operand: ArrayType,
	init_value: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's reduce operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be reduced.
	    init_value (ArrayType): The initial value for the reduction.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.reduce operation.
	"""
	if isinstance(operand, A4Q):
		operand = operand.materialize()
	if isinstance(init_value, A4Q):
		init_value = init_value.materialize()
	return lax.reduce(operand, init_value, *args, **kwargs)


@core.primitive_handler("mul")
def handle_mul(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's mul operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to multiply.
	    y (ArrayType): Second array to multiply.

	Returns:
	    The result of lax.mul operation.
	"""
	if isinstance(x, A4Q):
		x = x.materialize()
	if isinstance(y, A4Q):
		y = y.materialize()
	return lax.mul(x, y)


@core.primitive_handler("transpose")
def handle_transpose(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's transpose operation.

	Materializes A4Q input before performing the operation.
	Re-quantizes the result if the input was A4Q.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be transposed.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.transpose operation, potentially re-quantized.
	"""
	original_quantized = False
	if isinstance(operand, A4Q):
		operand = operand.materialize()
		original_quantized = True
	operand = lax.transpose(operand, *args, **kwargs)
	if original_quantized:
		operand = A4Q.quantize(operand, dtype=operand.dtype)
	return operand


@core.primitive_handler("conv_general_dilated")
def handle_conv(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's conv_general_dilated operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array (input).
	    rhs (ArrayType): Right-hand side array (kernel).
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.conv operation.
	"""
	if isinstance(lhs, A4Q):
		lhs = lhs.materialize()
	if isinstance(rhs, A4Q):
		rhs = rhs.materialize()
	return lax.conv_general_dilated(lhs, rhs, *args, **kwargs)


@core.primitive_handler("max")
def handle_max(
	primitive,
	x: ArrayType,
	y: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's max operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array for max comparison.
	    y (ArrayType): Second array for max comparison.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.max operation.
	"""
	if isinstance(x, A4Q):
		x = x.materialize()
	if isinstance(y, A4Q):
		y = y.materialize()
	return lax.max(x, y, *args, **kwargs)


@core.primitive_handler("exp")
def handle_exp(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's exp operation.

	Materializes A4Q input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply exponential to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.exp operation.
	"""
	if isinstance(x, A4Q):
		x = x.materialize()
	return lax.exp(x, *args, **kwargs)


@core.primitive_handler("log")
def handle_log(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's log operation.

	Materializes A4Q input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply logarithm to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.log operation.
	"""
	if isinstance(x, A4Q):
		x = x.materialize()
	return lax.log(x, *args, **kwargs)


@core.primitive_handler("reshape")
def handle_reshape(
	primitive: Primitive,
	operand: ArrayType,
	**kwargs: Any,
):
	"""
	Custom handler for JAX's reshape operation.

	This function handles reshaping for both regular arrays and A4Q quantized arrays.
	It materializes ArrayNF4 input before reshaping and re-quantizes the result if the input was ArrayNF4.

	Args:
	    primitive (Primitive): The JAX primitive being handled.
	    operand (ArrayType): The array to be reshaped.
	    new_sizes (Tuple[int, ...]): The desired new shape of the array.
	    dimensions (Tuple[int, ...], optional): The order in which dimensions should be permuted before reshaping.
	    **kwargs: Additional keyword arguments for the reshape operation.

	Returns:
	    ArrayType: The reshaped array, potentially re-quantized if the input was A4Q.

	Raises:
	    ValueError: If the new shape is not compatible with the original array's size.
	"""
	original_quantized = isinstance(operand, A4Q)

	if original_quantized:
		operand = operand.materialize()

	try:
		reshaped = lax.reshape(operand, **kwargs)
	except ValueError as e:
		raise ValueError(
			f"Reshape operation failed: {str(e)}. "
			f"Ensure the new shape {kwargs} is compatible with the original array size."
		) from e
	if original_quantized:
		reshaped = A4Q.quantize(reshaped, dtype=reshaped.dtype)
	return reshaped


@core.primitive_handler("concatenate")
def handle_concatenate(
	primitive,
	operands: Sequence[ArrayType],
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's concatenate operation.

	Materializes A4Q inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operands (Sequence[ArrayType]): Sequence of arrays to concatenate.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.concatenate operation.
	"""
	materialized_operands = [
		op.materialize() if isinstance(op, A4Q) else op for op in operands
	]
	return lax.concatenate(materialized_operands, *args, **kwargs)


@core.primitive_handler("convert_element_type")
def convert_element_type(
	primitive,
	arg: ArrayType,
	**params,
) -> ArrayType:
	"""Handle element type conversion for A4Q."""
	result = jax.tree_util.tree_map(
		partial(core.default_handler, primitive, **params), arg
	)
	result.dtype = params["new_dtype"]
	return result


@core.primitive_handler("broadcast_in_dim")
def handle_broadcast_in_dim(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle broadcast_in_dim for A4Q."""
	original_quantized = isinstance(operand, A4Q)
	array = operand
	if original_quantized:
		array = operand.materialize()
	result = jax.lax.broadcast_in_dim(array, *args, **kwargs)
	if original_quantized:
		result = A4Q.quantize(result, dtype=operand.dtype)
	return result


@core.primitive_handler("gather")
def handle_gather(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle gather for A4Q."""
	original_quantized = isinstance(operand, A4Q)
	array = operand
	if original_quantized:
		array = operand.materialize()
	result = jax.lax.gather(array, *args, **kwargs)
	return result



from fjformer.monitor.tracker import get_capacity_matrix as get_capacity_matrix
from fjformer.monitor.tracker import get_memory_information as get_memory_information
from fjformer.monitor.tracker import initialise_tracking as initialise_tracking
from fjformer.monitor.tracker import is_notebook as is_notebook
from fjformer.monitor.tracker import run as run
from fjformer.monitor.tracker import threaded_log as threaded_log



from functools import wraps
from typing import Any, Callable, Dict, List, Set, Tuple, Union

import jax
import jax.numpy as jnp
import optax
from jax import tree_util
from jax.dtypes import float0

from fjformer.core.implicit_array import implicit_compact
from fjformer.core.symbols import SymbolicConstant


# Custom exceptions
class OptimizerError(Exception):
	"""Base exception class for optimizer-related errors."""

	pass


class AxisError(OptimizerError):
	"""Exception raised for errors related to axis operations."""

	pass


class FreezingError(OptimizerError):
	"""Exception raised for errors during parameter freezing."""

	pass


class UpdateError(OptimizerError):
	"""Exception raised for errors during parameter updates."""

	pass


def vmap_all_but_one(f: Callable, axis: int, out_ndim: int = 0) -> Callable:
	"""
	Repeatedly calls vmap to map over all axes except for `axis`.
	All args will be mapped on the same dimensions.

	Args:
	    f: The function to be mapped.
	    axis: The axis to exclude from mapping.
	    out_ndim: The number of dimensions in the output.

	Returns:
	    A wrapped function that applies vmap to all axes except the specified one.

	Raises:
	    AxisError: If the specified axis is out of bounds.
	"""

	@wraps(f)
	def inner(*args: Any) -> Any:
		n_dim = args[0].ndim
		if axis >= n_dim:
			raise AxisError(f"Axis {axis} is out of bounds for array of dimension {n_dim}")
		fn = f
		vmap_dim = 1
		out_dim = out_ndim
		for i in reversed(range(n_dim)):
			if i == axis:
				vmap_dim = 0
				out_dim = 0
			else:
				fn = jax.vmap(fn, vmap_dim, out_dim)
		return fn(*args)

	return inner


def freeze_subtrees(
	optimizer: optax.GradientTransformation,
	label_fn: Callable[[Any], Dict[str, Any]],
	use_scalar_zeros: bool = False,
) -> optax.GradientTransformation:
	"""
	Wraps an optimizer such that subtrees specified by label_fn will receive zeros as updates.

	Args:
	    optimizer: The original optimizer to be wrapped.
	    label_fn: A function that labels subtrees as "freeze" or "train".
	    use_scalar_zeros: If True, use scalar zeros instead of array zeros.

	Returns:
	    A wrapped optimizer that freezes specified subtrees.

	Raises:
	    FreezingError: If there's an error during the freezing process.
	"""
	try:
		multi_transformed_optimizer = optax.multi_transform(
			{
				"freeze": (set_to_zero_scalar() if use_scalar_zeros else optax.set_to_zero()),
				"train": optimizer,
			},
			label_fn,
		)

		def new_update(grads: Any, opt_state: Any, params: Any) -> Tuple[Any, Any]:
			def map_float0(param: Any, grad: Any) -> Any:
				if grad.dtype == float0:
					return (
						jnp.zeros((), param.dtype) if use_scalar_zeros else jnp.zeros_like(param)
					)
				return grad

			fixed_grads = jax.tree_util.tree_map(map_float0, params, grads)
			return multi_transformed_optimizer.update(fixed_grads, opt_state, params)

		return optax.GradientTransformation(multi_transformed_optimizer.init, new_update)
	except Exception as e:
		raise FreezingError(f"Error in freeze_subtrees: {str(e)}") from e


def freeze_keys(
	optimizer: optax.GradientTransformation,
	arr_type: type,
	keys: Union[List[str], Set[str]],
	use_scalar_zeros: bool = False,
) -> optax.GradientTransformation:
	"""
	Freezes specific keys in the optimizer.

	Args:
	    optimizer: The original optimizer to be wrapped.
	    arr_type: The type of array to be processed.
	    keys: A list or set of keys to be frozen.
	    use_scalar_zeros: If True, use scalar zeros instead of array zeros.

	Returns:
	    A wrapped optimizer with specified keys frozen.

	Raises:
	    FreezingError: If there's an error during the freezing process.
	"""
	try:
		keys_set = set(keys)

		def label_leaf(leaf: Any) -> Union[str, Any]:
			if not isinstance(leaf, arr_type):
				return "train"

			children, aux_data = leaf.tree_flatten_with_keys()
			labels = ["freeze" if key in keys_set else "train" for key, _ in children]
			struct = leaf.tree_unflatten(aux_data, labels)
			return struct

		def label_fn(root: Any) -> Any:
			return jax.tree_util.tree_map(
				label_leaf, root, is_leaf=lambda x: isinstance(x, arr_type)
			)

		return freeze_subtrees(optimizer, label_fn, use_scalar_zeros=use_scalar_zeros)
	except Exception as e:
		raise FreezingError(f"Error in freeze_keys: {str(e)}") from e


def apply_updates(params: optax.Params, updates: optax.Updates) -> optax.Params:
	"""
	Applies updates to parameters, supporting SymbolicConstant instances.

	Args:
	    params: The current parameters.
	    updates: The updates to be applied.

	Returns:
	    The updated parameters.

	Raises:
	    UpdateError: If there's an error during the update process.
	"""
	try:
		updates_flat, update_struct = tree_util.tree_flatten(
			updates, is_leaf=lambda x: isinstance(x, SymbolicConstant)
		)
		semi_flat_params = update_struct.flatten_up_to(params)

		updated_flat = implicit_compact(optax.apply_updates)(semi_flat_params, updates_flat)
		updated = update_struct.unflatten(updated_flat)
		return updated
	except Exception as e:
		raise UpdateError(f"Error in apply_updates: {str(e)}") from e


def set_to_zero_scalar() -> optax.GradientTransformation:
	"""
	Returns a gradient transformation that sets all gradients to 0 to make downstream constant folding cheaper.

	Returns:
	    A gradient transformation that sets all gradients to scalar zeros.

	Raises:
	    OptimizerError: If there's an error during the transformation process.
	"""
	try:

		def init_fn(params: Any) -> optax.EmptyState:
			del params
			return optax.EmptyState()

		def update_fn(updates: Any, state: Any, params: Any = None) -> Tuple[Any, Any]:
			return (
				jax.tree_util.tree_map(lambda x: jnp.zeros((), x.dtype), updates),
				state,
			)

		return optax.GradientTransformation(init_fn, update_fn)
	except Exception as e:
		raise OptimizerError(f"Error in set_to_zero_scalar: {str(e)}") from e



from fjformer.lora.rapture import LoraRapture as LoraRapture
from fjformer.lora.rapture import LoraWeight as LoraWeight
from fjformer.lora.rapture import RaptureConfig as RaptureConfig
from fjformer.lora.rapture import RaptureModule as RaptureModule
